/*
 * This module hooks into the netfilter infrastructure to intercept
 * network packets during their journey through the kernel.
 *
 * Interesting background reading material (even though a bit dated):
 * http://phrack.org/issues/61/13.html
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/list.h>
#include <linux/udp.h>
#include <linux/tcp.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/ipv6.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <net/ipv6.h>

MODULE_AUTHOR("Juergen Schoenwaelder");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Simple netfilter kernel module");

#define kwarn(FMT, ...) \
        printk(KERN_WARNING "[%s] "FMT"\n", __this_module.name, ##__VA_ARGS__)

#define kinfo(FMT, ...) \
        printk(KERN_INFO "[%s] "FMT"\n", __this_module.name, ##__VA_ARGS__)

/*
 * Hook function for processing outgoing IP packets.
 */

static unsigned int
hook_func_out(unsigned int hooknum, struct sk_buff *skb,
	      const struct net_device *in, const struct net_device *out,
	      int (*okfn)(struct sk_buff *))
{
    return NF_ACCEPT;
}

/*
 * Hook function for processing incoming IP packets.
 */

static unsigned int
hook_func_in(unsigned int hooknum,
	     struct sk_buff *skb,
	     const struct net_device *in,
	     const struct net_device *out,
	     int (*okfn)(struct sk_buff *))
{
    unsigned char *nw_header;
    
    struct iphdr *ip4_header = NULL;
    struct ipv6hdr *ip6_header = NULL;
    struct udphdr *udp_header = NULL;
    struct tcphdr *tcp_header = NULL;

    unsigned int src_port = 0;
    unsigned int dst_port = 0;

    nw_header = skb_network_header(skb);
    if (!nw_header || !in) {
	return NF_ACCEPT;
    }

    switch (nw_header[0] & 0xf0) {
    case 0x40:
	// kinfo("ipv4 packet from interface '%s'", in->name);
	ip4_header = (struct iphdr *) skb_network_header(skb);
	break;
    case 0x60:
	// kinfo("ipv6 packet from interface '%s'", in->name);
	ip6_header = (struct ipv6hdr *) skb_network_header(skb);
	break;
    default:
	return NF_ACCEPT;
    }

    if (ip4_header) {
	/*
	 * The IPv4 layer has not seen the packet yet and hence
	 * skb_transport_header() does not provide the expected
	 * result.
	 */
	switch (ip4_header->protocol) {
	case 6: /* any suitable #defines? */
	    tcp_header = (struct tcphdr *) (skb_network_header(skb)
					    + ip4_header->ihl * 4);
	    break;
	case 17: /* any suitable #defines? */
	    udp_header = (struct udphdr *) (skb_network_header(skb)
					    + ip4_header->ihl * 4);
	    break;
	default:
	    return NF_ACCEPT;
	}
    }
    
    if (ip6_header){
	/*
	 * Interestingly, skb_transport_header() seems to give us
	 * a pointer to the transport header for IPv6...
	 */
	switch (ip6_header->nexthdr) {
	case NEXTHDR_TCP:
	    tcp_header = (struct tcphdr *) (skb_transport_header(skb));
	    break;
	case NEXTHDR_UDP:
	    udp_header = (struct udphdr *) (skb_transport_header(skb));
	    break;
	default:
	    return NF_ACCEPT;
	}
    }
    
    if (udp_header) {
	src_port = (unsigned int) ntohs(udp_header->source);
	dst_port = (unsigned int) ntohs(udp_header->dest);
    }
    
    if (tcp_header) {
	src_port = (unsigned int) ntohs(tcp_header->source);
	dst_port = (unsigned int) ntohs(tcp_header->dest);
    }

    return NF_ACCEPT;
}


/*
 * Module initialization and cleanup code...
 */

static struct nf_hook_ops nfhooks[] = {
    {
	.owner = THIS_MODULE,
	.hook = hook_func_in,
	.hooknum = NF_INET_LOCAL_IN,
	.pf = PF_INET,
	.priority = NF_IP_PRI_FIRST,
    },
    {
	.owner = THIS_MODULE,
	.hook = hook_func_out,
	.hooknum = NF_INET_LOCAL_OUT,
	.pf = PF_INET,
	.priority = NF_IP_PRI_FIRST,
    },
    {
	.owner = THIS_MODULE,
	.hook = hook_func_in,
	.hooknum = NF_INET_LOCAL_IN,
	.pf = PF_INET6,
	.priority = NF_IP_PRI_FIRST,
    },
    {
	.owner = THIS_MODULE,
	.hook = hook_func_out,
	.hooknum = NF_INET_LOCAL_OUT,
	.pf = PF_INET6,
	.priority = NF_IP_PRI_FIRST,
    }
};

static int __init
nf_init(void)
{
    int i;

    for (i = 0; i < sizeof(nfhooks)/sizeof(nfhooks[0]); i++) {
	nf_register_hook(&nfhooks[i]);
    }
    return 0;
}

static void __exit
nf_exit(void)
{
    int i;
    
    for (i = 0; i < sizeof(nfhooks)/sizeof(nfhooks[0]); i++) {
	nf_unregister_hook(&nfhooks[i]);
    }
}

module_init(nf_init);
module_exit(nf_exit);
